// Copyright (C) 1998,2000 Logi Ragnarsson

package org.logi.crypto.modes;
import org.logi.crypto.*;
import org.logi.crypto.keys.*;

import java.util.Random;

/**
 * Use this class to decrypt ciphertext generated by EncryptOFB.<p>

 * Note that unlike ECB or CBC mode, OFB mode uses teh <i>same</i>
 * key to encrypt and decrypt, even with asymmetric ciphers.

 * @author <a href="http://www.logi.org/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@logi.org">logir@logi.org</a>)
 *
 * @see org.logi.crypto.modes.EncryptOFB
 */
public class DecryptOFB extends DecryptMode{
    
   /** Calculates the stream. */
   private OFBProducer producer;
   
   /* Stores desired buffer size until setKey() is called. */
   private int bufSize;
   
   // keyStream[keyPos..keyStream.length-1] is the next unused key-stream
   // bytes.
   private byte[] keyStream;
   private int keyPos;
   
   /* 
    * If the IV has not been read and the thread not started, then IV will
    * not be null and IV[0..IVPos-1] will contain the bytes we have so
    * far. */
   private byte[] IV;
   private int IVPos=0;

   private CipherKey key;
   
   /**
    * Create a new OFB-mode decrypt session with the specified
    * <code>key</code>. A buffer of <code>bufSize</code> bytes is created
    * to hold a pre-calculated xor-stream.
    */
   public DecryptOFB(CipherKey key, int bufSize) {
      this.bufSize=bufSize;
      setKey(key);
   }
    
   /**
    * Create a new OFB-mode decrypt session with no key. No
    * encryption can be performed until the <code>setKey()</code>
    * method has been called.
    * <p>
    * A buffer of <code>bufSize</code> bytes is created to hold a
    * pre-calculated xor-stream.*/
   public DecryptOFB(int bufSize){
      this.bufSize=bufSize;
   }
    
   /** Return the key used for encryption. */
   public CipherKey getKey(){
      return key;
   }
   
   /**
    * Set the key to use for encryption. The key can only be set once in
    * this version of the library. The 1.1.x and eventually 1.2.x series
    * allows dynamic re-keying. */
   public synchronized void setKey(CipherKey key) {
      if(producer!=null){
	 producer.kill();
	 producer=null;
      }
      this.key = key;
      IV = new byte[key.plainBlockSize()];
   }
    
   /**
    * Return the size of the blocks of plaintext output by this object.
    */
   public int plainBlockSize(){
      return 1;
   }
   
   /**
    * Send bytes to the DecryptOFB object for encryption.
    * <p>
    * Decrypt <code>length</code> bytes from <code>source</code>,
    * starting at <code>i</code> and return the plaintext.
    */
    public synchronized byte[] decrypt(byte[] source, int i, int length){
       if (IV!=null){
	  // we haven't completed reading the IV yet.
	  if(IVPos==0 && length>=IV.length) {
	     // we can just start the producer directly from source
	     producer=new OFBProducer(key, bufSize, source,i);
	     length -= producer.ibs;
	     i += producer.ibs;
	     IV=null;
	  } else {
	     // we must collect the IV from pieces and then start producer
	     int get = Math.min(IV.length-IVPos, length);
	     System.arraycopy(source,i, IV,IVPos, get);
	     IVPos+=get;
	     if(IVPos==IV.length) {
		producer=new OFBProducer(key, bufSize, IV,0);
		IV=null;
	     }
	     length-=get;
	     if(length==0) {
		// there is nothing but (possibly partial) IV
		return new byte[0];
	     }
	     i+=get;
	  }
       }
       
       byte[] dest = new byte[length];
       int destPos=0;
       
       if(keyStream==null) {
	  keyStream = producer.nextBuffer();
	  keyPos = 0;
       }
       
       while(length>0){
	  if(keyPos==keyStream.length) {
	     keyStream = producer.nextBuffer();
	     keyPos = 0;
	  }
	  int n = Math.min(length, keyStream.length-keyPos);
	  length-=n;
	  int end=i+n;
	  while(i<end){
	     dest[destPos++] = (byte)(source[i++] ^ keyStream[keyPos++]);
	  }
       }
       
       return dest;
    }
   
    /**
     * Close files and kill threads owned by the object. This should
     * be called to make sure all resources are freed.
     */
    public void close(){
        if(producer==null)
            return;
        producer.kill();
        producer=null;
    }
    
}
